@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1995, 2017, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2013, 2021, MariaDB Corporation.
+Copyright (c) 2013, 2022, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -229,8 +229,6 @@ fail:
 
   /* Remove doublewrite pages from LRU */
   buf_pool_invalidate();
-
-  ib::info() << "Doublewrite buffer created";
   goto start_again;
 }
 
@@ -342,7 +340,7 @@ func_exit:
 /** Process and remove the double write buffer pages for all tablespaces. */
 void buf_dblwr_t::recover()
 {
-  ut_ad(recv_sys.parse_start_lsn);
+  ut_ad(log_sys.last_checkpoint_lsn);
   if (!is_initialised())
     return;
 
@@ -360,13 +358,13 @@ void buf_dblwr_t::recover()
       continue;
 
     const lsn_t lsn= mach_read_from_8(page + FIL_PAGE_LSN);
-    if (recv_sys.parse_start_lsn > lsn)
+    if (log_sys.last_checkpoint_lsn > lsn)
       /* Pages written before the checkpoint are not useful for recovery. */
       continue;
     const uint32_t space_id= page_get_space_id(page);
     const page_id_t page_id(space_id, page_no);
 
-    if (recv_sys.scanned_lsn < lsn)
+    if (recv_sys.lsn < lsn)
     {
       ib::info() << "Ignoring a doublewrite copy of page " << page_id
                  << " with future log sequence number " << lsn;
